"""Pip requirements file parsing.

Specification is defined within pip itself and documented under:

- https://pip.pypa.io/en/stable/reference/pip_install/#requirements-file-format
- https://github.com/pypa/pip/blob/master/src/pip/_internal/req/constructors.py#L291

"""

from __future__ import annotations
